classdef SnowPack
    properties
        %sizes of vector/matrix/array for the simulation. An xy plane
        %is one 'layer'. The z-direction is vertical height
        x_size;
        y_size;
        z_size;
        kc;
        zp;
        %Vector holding the simulation data. x_size x y_size x z_size
        data;
        xx;
        yy;
        zz;
        x_distance_max;
        y_distance_max;
        z_distance_max;
        distance_interval_xy;
        distance_interval_z
        time_interval;
        N = 100;
        alpha = .2;
        temperature_profile;
        temp_min;
        temp_max;
        water_g_mol = 16;
        water_diffusivity = .282 / 100 ^ 2; % m^2/s at 25 degC, converted from cm^2/s
        layer_volume;
        kn_vector;
        phin_vector;
        k_vals;
        h = figure;
        %Empyrically derived rate from papers, conforming to a layer
        %sublimating over the course of 24 hours.
        sublimation_particle_rate = 20;
    end
    methods
        % Initialize the simulation data
        % -Set the sizes of the data matrix
        % -Create the meshgrid holding the real coordinates to create r
        %  vectors
        % -Set the parameters given in the function call
        function obj = initialize(obj, ...
                x_size, y_size, z_size, ...
                x_distance_max, y_distance_max, z_distance_max, ...
                kc, zp, ...
                temp_min, temp_max)
            obj.x_size = x_size;
            obj.y_size = y_size;
            obj.z_size = z_size;
            obj.x_distance_max = x_distance_max;
            obj.y_distance_max = y_distance_max;
            obj.z_distance_max = z_distance_max;
            obj.kc = kc;
            obj.zp = zp;
            obj.data = zeros(obj.x_size, obj.y_size, obj.z_size);
            obj.temp_min = temp_min;
            obj.temp_max = temp_max;
            z_num_intervals = obj.z_size;
            obj.temperature_profile = temp_max:-abs(temp_max - temp_min) / z_num_intervals: temp_min;
            % Change to kelvin
            obj.temperature_profile = obj.temperature_profile + 273.15;
            % Compute the volume of one layer, xy plane, layer of z
            
            obj.layer_volume = (obj.x_distance_max) * ...
                               (obj.y_distance_max) * ...
                               (obj.z_distance_max / obj.z_size);
            obj.kn_vector = zeros(obj.N, 3);
            obj.phin_vector = zeros(obj.N, 1);
            [obj.xx, obj.yy, obj.zz] = meshgrid(0:x_distance_max/x_size:x_distance_max,...
                                                0:y_distance_max/y_size:y_distance_max,...
                                                0:z_distance_max/z_size:z_distance_max);
        end
        function obj = initialize_snowpack(obj)
            for k=1:obj.z_size
                disp('Z index computed: ');
                disp(k);
                for j=1:obj.y_size
                    for i=1:obj.x_size
                        r = [obj.xx(i, j, k) obj.yy(i, j, k) obj.zz(i, j, k)];
                        entry_value = calculate_coordinate(r, obj.N, obj.zp, obj.kc, obj.alpha, obj.kn_vector, obj.phin_vector);
                        obj.data(i, j, k) = entry_value;
                    end
                end
            end
        end
        function obj = plot_snowpack(obj)
            imagesc(obj.data(:, :, 1));
            colormap(flipud(gray));
            set(gca,'YDir','normal');
        end
        function obj = plot_snowpack_layers(obj, ind)
            clf;
            for i=1:obj.z_size
                subplot(4, 3, i);
                imagesc(obj.data(:, :, i));
                colormap(flipud(gray));
                set(gca,'YDir','normal');
                xlabel('X Distance in (cm)');
                ylabel('Y Distance in (cm)');
                plot_title = strcat('XY Plane Cut of Snowpack at Z=', num2str(obj.zz(1, 1, i) * 10 ^ 2, 2), 'cm');
                title(plot_title);
                tick_distance = obj.x_distance_max * 100;
                set(gca,'XTickLabel',[tick_distance/5:tick_distance/5:tick_distance]);
                set(gca,'YTickLabel',[tick_distance/5:tick_distance/5:tick_distance]);
                subplot(4, 3, 1);
                if ind ~= -1
                    text(70,70,strcat('Snow Pack Time Evolution T=', num2str(ind)));
                end
            end
        end
        function obj = plot_snowpack_walkthrough(obj)
            clf;
            for i=1:obj.z_size
                imagesc(obj.data(:, :, i));
                colormap(flipud(gray));
                set(gca,'YDir','normal');
                xlabel('X Distance in (cm)');
                ylabel('Y Distance in (cm)');
                plot_title = strcat('XY Plane Cut of Snowpack at Z=', num2str(obj.zz(1, 1, i) * 10 ^ 2, 2), 'cm');
                title(plot_title);
                tick_distance = obj.x_distance_max * 100;
                set(gca,'XTickLabel',[tick_distance/5:tick_distance/5:tick_distance]);
                set(gca,'YTickLabel',[tick_distance/5:tick_distance/5:tick_distance]);
                saveas(obj.h, strcat('xy-cut', num2str(i)), 'jpg');
            end
        end
        function obj = save_data(obj, filename)
            dlmwrite(filename, obj.data);
        end
        function obj = init_random_vectors(obj)
            obj.kc = obj.kc / 32;
            a = obj.zp + 1;
            b = obj.kc / (obj.zp + 1);
            pd = makedist('Gamma', 'a', a, 'b', b);
            obj.k_vals = random(pd, obj.N, 1);
            %Compute the summation in the bi-continuous model
            for i = 1:obj.N
                % Create random phase
                obj.phin_vector(i) = rand * 2 * pi;
                % Create k random direction vector with k magnitude
                obj.kn_vector(i, :) = create_k_vector(obj.k_vals(i));
            end
        end
        % Compute the fick flow for the ith layer. i=1 corresponds to the
        % max temperature
        function J = compute_fick_flow(obj, i)
            d_phi_1 = obj.compute_mol_water(i);
            d_phi_2 = obj.compute_mol_water(i+1);
            d_phi = d_phi_1 - d_phi_2;
            dx = obj.z_distance_max / obj.z_size;
            d_phi_dx = d_phi / dx;
            J = -obj.water_diffusivity * d_phi_dx;
        end
        function moles = compute_mol_water(obj, i)
            % Compute the water vapor pressure for the current layer's
            % temperature
            layer_vapor_pressure = clausius_clapeyron_pressure(obj.temperature_profile(i) - 273.15);
            % Compute the moles of water using the ideal gas law
            moles = ideal_gas_law_n_particles(layer_vapor_pressure, obj.layer_volume, obj.temperature_profile(i));
        end
        function mass = compute_water_mass(obj, i)
            moles = obj.compute_mol_water(i);
            mass = moles * 18 / 10 ^ 3; %Mass of water in kg.
        end
        function obj = evolve_snowpack(obj)
            new_data = zeros(obj.x_size, obj.y_size, obj.z_size);
            for k=1:obj.z_size
                new_layer = obj.evolve_layer(k, obj.data(:, :, k));
                %disp('new_layer:');disp(size(new_layer));
                new_data(:, :, k) = new_layer;
            end
            obj.data = new_data;
        end
        function obj = time_evolve_snowpack(obj, steps)
            for s=1:steps
                obj.plot_snowpack_layers(s - 1);
                %saveas(obj.h, strcat('metmorphism', num2str(s)), 'jpg');
                disp('paused');
                pause();
                disp('time step: ');
                disp(s);
                obj = obj.evolve_snowpack();
            end
        end
        function new_layer = sublimate_layer(obj, k, original_layer)
            particles = obj.particles_at_layer(k);
            num_particles = size(particles, 2);
            frequencies = zeros(num_particles, 1);
            layer = original_layer;
            for n=1:num_particles
                frequencies(n) = obj.particle_sublimation_frequency(particles(:, n), k);
            end
            frequencies = frequencies / sum(frequencies);
            chosen_particles = sample_from_vector(frequencies', obj.sublimation_particle_rate);
            for n=1:size(chosen_particles, 1)
                p = particles(:, chosen_particles(n));
                layer(p(1), p(2)) = 0;
            end
            new_layer = layer;
        end
        function new_layer = deposit_layer(obj, k, original_layer)
            frequencies = zeros(obj.x_size * obj.y_size);
            for i=1:obj.x_size
                for j=1:obj.y_size
                    if obj.data(i, j, k) ~= 1
                        frequencies(i + (j - 1) * obj.y_size) = 1 + obj.neighbor_particle_count([i j], k);
                    else
                        frequencies(i + (j - 1) * obj.y_size) = 0;
                    end
                end
            end
            frequencies = frequencies / sum(frequencies);
            chosen_locations = sample_from_vector(frequencies', obj.sublimation_particle_rate);
            layer = original_layer;
            for n=1:size(chosen_locations)
                y = obj.periodic_bounds(floor(chosen_locations(n) / obj.x_size) + 1);
                x = obj.periodic_bounds(mod(chosen_locations(n), obj.x_size) + 1);
                layer(x, y) = 1;
            end
            new_layer = layer;
        end
        function new_layer = evolve_layer(obj, k, original_layer)
            layer = obj.sublimate_layer(k, original_layer);
            layer = obj.deposit_layer(k, layer);
            layer = obj.deposit_layer(k, layer);
            new_layer = layer;
        end
        function f = particle_sublimation_frequency(obj, particle, k)
            f = 1 / (obj.neighbor_particle_count(particle, k) + 1);
        end
        function nb_count = neighbor_particle_count(obj, particle, k)
            x = particle(1);
            y = particle(2);
            pxl = obj.periodic_bounds(x - 1);
            pxm = x;
            pxr = obj.periodic_bounds(x + 1);
            pyb = obj.periodic_bounds(y - 1);
            pym = y;
            pyu = obj.periodic_bounds(y + 1);
            nb_count = obj.data(pxl, pyb, k) +...
                obj.data(pxl, pym, k) +...
                obj.data(pxl, pyu, k) +...
                obj.data(pxm, pyb, k) +...
                obj.data(pxm, pyu, k) +...
                obj.data(pxr, pyu, k) +...
                obj.data(pxr, pym, k) +...
                obj.data(pxr, pyb, k);
        end
        function b = periodic_bounds(obj, n)
            if n == 0
                b = obj.x_size;
            elseif n == obj.x_size + 1
                b = 1;
            else
                b = n;
            end
        end
        function particles = particles_at_layer(obj, k)
            particles = zeros(2, 0);
            for i=1:obj.x_size
                for j=1:obj.y_size
                    if obj.data(i, j, k) == 1
                        particles(:, size(particles, 2) + 1) = [i j];
                    end
                end
            end
        end
    end
end